var /* BigDecimal */ xmin, dx, yval;
var /* int */ columnCount, rowNumber, maxIterations;
var /* boolean */ highPrecision;
var /* double */ escrad2; // used by hiprec only
var show_dem = false;

var /* int */ rowNum, batchID, workerNumber;
var /* int */ colStart, colStride, pxWidth, pxHeight, prio;
var /* int[] */ xprios; // input array: tells which pixels to evaluate

var /* int[] */ rowIts;
var /* double[] */ rowDEM;

var ArrayType = this.Float64Array || Array; // this.Uint32Array || Array;

var pingTO = null;

// ------- support for high-precision calculation ------------

var FRCM1MASK, FRACMASK, SIGNMASK, CARRYMASK, IAradix, IAradixBits, IArecip;
IAradixBits = 26;
if (IAradixBits == 16) {
  FRCM1MASK = 0x7FFF;
  FRACMASK = 0xFFFF;
  SIGNMASK = 0x8000;
  CARRYMASK = 0x10000;
  IAradix = 65536.0;
} else if (IAradixBits == 20) {
  FRCM1MASK = 0x7FFFF;
  FRACMASK = 0xFFFFF;
  SIGNMASK = 0x80000;
  CARRYMASK = 0x100000;
  IAradix = 1048576.0;
} else {
  FRCM1MASK = 0x1FFFFFF;
  FRACMASK = 0x3FFFFFF;
  SIGNMASK = 0x2000000;
  CARRYMASK = 0x4000000;
  IAradix = 67108864.0;
}
IArecip = 1.0/IAradix;

var  /* int[][] */ xs;
var /* int[] */ y;

var /* int[] */ multmp;
var /* int[] */ qzr2, qzi2, qrad2;
var /* int[] */ qzr, qzi, qzit;

var /* int */ chunks;

var /* double */ log2of10 = Math.log(10)/Math.log(2);

/* Create arrays and lookup tables used for the high-precision
   iteration calculations. The main task is to create the two-dimensional
   array xs[][], each row of which is the extended-precision representation
   of the x (real) value for a column of pixels. */
function createHPData()
{
  chunks = xmin.length - 1;
  y = new ArrayType(chunks+1);
  xs = new Array(columnCount);
  xs[0] = xmin; // must have xs.length = chunks+1
  for (var i = 1; i < columnCount; i++) {
    xs[i] = new ArrayType(chunks+1);
  }
  if (columnCount > 1) {
    /* xs[0][] was set equal to xmin[] above. Now we iteratively add dx[]
       to xs[i-1][] to get xs[i][] */
    for (var i = 1; i < columnCount; i++) {
      /* Copy xs[i-1][] to xs[i][] */
      for (var j = 0; j <= chunks; j++) {
        xs[i][j] = xs[i-1][j];
      }
      /* Add dx[] to xs[i][] */
      IAadd(xs[i],dx,chunks+1);
    }
  }
  qzr = new ArrayType(chunks);
  qzi = new ArrayType(chunks);
  qzr2 = new ArrayType(chunks);
  qzi2 = new ArrayType(chunks);
  qzit = new ArrayType(chunks);
  qrad2 = new ArrayType(chunks);
  multmp = new ArrayType(chunks);
} /* End of create.HPData */

onmessage = function(msg) {
  var data = msg.data;
  var rv;
  if ( data[0] == "mw_setup" ) { 
    batchID = data[1];
    maxIterations = data[2];
    highPrecision = data[3];
    workerNumber = data[4];
    escrad2 = data[5];
    if (pingTO == null) {
      pingTO = setTimeout(sendPing, 1000);
    }
    shout("got "+data[0]);
  } else if ( data[0] == "mw_iter" ) {
    var dt1 = data[1];
    rowNum = dt1[0];  // j.row
    shout("got "+data[0]+" for row "+rowNum);
    colStart = dt1[1];
    colStride = dt1[2];
    pxWidth = dt1[3];
    pxHeight = dt1[4];
    prio = dt1[5];
    columnCount = data[2]; // j.columns
    xprios = data[6];// prios for each pixel
    rowIts = new Array(columnCount);
    rowDEM = new Array(columnCount);
    var j=0;
    if (highPrecision) {
      xmin = data[3];  // j.xmin (IntArray pointer only, we'll modify it)
      dx = data[4];    // j.dx
      yval = data[5];  // j.yVal
      createHPData();  // creates xs[][] containing x values for each pixel
      for (var i=colStart; i<columnCount; i+=colStride) {
        /* We pass in the real coordinate xs[i] which was pre-computed
           by the create.HPData function, and the imaginary coordinate yval
           that was given to us by the work request. */
        if (prio > xprios[i]) {
          rv = dwellHP(xs[i], yval);
          rowIts[j] = rv[0];
          rowDEM[j] = rv[1];
        }
        j++;
      }
    } else {
      var xmin_d = data[3];
      var dx_d = data[4];
      var yval_d = data[5];
      for (var i=colStart; i<columnCount; i+=colStride) {
        if (prio > xprios[i]) {
          rv = dwell53(xmin_d + dx_d*i, yval_d);
          rowIts[j] = rv[0];
          rowDEM[j] = rv[1];
        }
        j++;
      }
    }
    var returnData = [ batchID,
                       [rowNum, colStart, colStride, pxWidth, pxHeight, prio],
                       [rowIts, rowDEM], workerNumber ];
    postMessage(returnData);
    /* message is received by function jobFinished(msg) in index.html */
  }
} /* End of function on.message */

function sendPing() {
  var returnData = [-1, [0, 0, 0, 0, 0, 0], [null, null], workerNumber];
  postMessage(returnData);
  pingTO = setTimeout(sendPing, 1000);
}

var max_shouts = 3;
var num_shouts = 0;
var dbglog = "";
function shout(msg) {
  if (num_shouts < max_shouts) {
    dbglog = dbglog + " // " + msg;
  } else if (num_shouts == max_shouts) {
    var returnData = [-2, num_shouts, dbglog, workerNumber];
    postMessage(returnData);
  }
  num_shouts++;
}

/* dwell */
function dwell53( /* double */ r, /* double */ i)
{
  var count = 0;
  var zr = r;
  var zi = i;
  var zr2 = zr*zr;
  var zi2 = zi*zi;
  var dzr = 1.0;
  var dzi = 0.0;
  var rad2 = 0.0;
  var gg = true;
  while ((count < maxIterations) && gg) {
    rad2 = 2.0*(zr*dzr - zi*dzi) + 1.0;
    dzi = 2.0*(zi*dzr + zr*dzi);
    dzr = rad2;

    /* var new_zr = zr*zr - zi*zi + r; */
    zi = 2.0*zr*zi + i;
    zr = zr2 - zi2 + r;
    zr2 = zr*zr;
    zi2 = zi*zi;
    rad2 = zr2 + zi2;
    count++;
    if (rad2 > escrad2)
      gg = false;
  }
  if (count < maxIterations) {
    /* escaped case */
    zr2 = (dzr*dzr)+(dzi*dzi);
    if (zr2 > 0.0) {
      zi2 = 2.0 * Math.log(Math.sqrt(rad2)) * Math.sqrt(rad2/zr2);
    } else {
      /* rare error case: avoid sqrt(negative) */
      zi2 = 1.0;
    }
    if (show_dem) { // if (show_dem)
      if (zi < 0.0) {
        count++;
      }
    }
    return [count, zi2];
  }
  /* else, not escaped */
  return [-1 , 0.0];
} /* End of dwell.53 */

/* Convert an Int[] extended-precision number to 64-bit floating point,
   with truncation. This is used only for the DEM/M derivative calculation,
   where float64 precision is more than adequate.
     WARNING: The input array must have at least 4 'chunks' */
function IA_f64( /* Int[] */ x, /* int */ chunks )
{
  var rv = 0.0;
  var k2pn16 = IArecip; // 1.0/65536.0
  var bp = 0;
  var scale = 1.0;
  while ((x[bp] == 0)
      && (x[bp+1] < SIGNMASK) // 0x8000
      && (bp+4 < chunks)
  ) {
    scale *= k2pn16;
    bp++;
  }
  if (bp == 0) {
    while ((x[bp] == FRACMASK) // 0xFFFF
       && (x[bp+1] > FRCM1MASK) // 0x7FFF
        && (bp+4 < chunks)) {
      scale *= k2pn16;
      bp++;
    }
  }
  var x0 = x[bp];
  var x1 = x[bp+1];
  var x2 = x[bp+2];
  var x3 = x[bp+3];
  var neg = (x[bp] & SIGNMASK) != 0; // 0x8000
  if (neg) {
    /* Complement */
    x0 = FRACMASK-x0; // 0xFFFF
    x1 = FRACMASK-x1; // 0xFFFF
    x2 = FRACMASK-x2; // 0xFFFF
    x3 = FRACMASK-x3; // 0xFFFF
    /* add 1, carry propagation */
    x3 += 1;
    if (x3 > FRACMASK) { // 0xFFFF
      x3 = 0; x2 += 1;
      if (x2 > FRACMASK) { // 0xFFFF
        x2 = 0; x1 += 1;
        if (x1 > FRACMASK) { // 0xFFFF
          x1 = 0; x0 = (x0 + 1) & FRACMASK; // 0xFFFF
        }
      }
    }
    scale *= -1.0;
  }
  rv = scale * (x0 + (x1 + (x2+x3*IArecip)*IArecip)*IArecip); // .../65536.0
  return rv;
} /* End of IA.f64 */

/* Test an Int[] extended-precision number to see if it exceeds the
   square of the "escape radius". Currently it tests if it equals
   or exceeds 8.0 so the effective escape radius is sqrt(8) */
function IA_escaped( qr0, qr1 )
{
  var r2 = 1.0*qr0+(qr1*IArecip); // qr1/65536.0
  return (r2 > escrad2);
}

function IAcopy(destArray, destStart, sourceArray, sourceStart, count)
{
  for (var i = 0; i < count; i++) {
    destArray[destStart + i] = sourceArray[sourceStart + i];
  }
} /* End of IA.copy */

function IAnegate( /* int[] */ x, /* int */ chunks) {
  for (var i = 0; i < chunks; i++)
    x[i] = FRACMASK-x[i]; // 0xFFFF
  ++x[chunks-1];
  for (var i = chunks-1;
       (i > 0) && ((x[i] & CARRYMASK) != 0) ; // 0x10000
       i--)
  {
    x[i] &= FRACMASK; // 0xFFFF
    ++x[i-1];
  }
  x[0] &= FRACMASK; // 0xFFFF
} // End of IA.negate

function IAadd( /* int[] */ x, /* int[] */ dx, /* int */ count) {
  var carry = 0;
  for (var i = count - 1; i >= 0; i--) {
    x[i] += dx[i];
    x[i] += carry;
    carry = (x[i] * IArecip) & 3; // x[i] >>> 16;
    x[i] &= FRACMASK; // 0xFFFF
  }
} // end of IA.add

function IAmultiply( /* int[] */ x, /* int[] */ y, /* int */ count){  // Can't allow x == y !
  var neg1 = (x[0] & SIGNMASK) != 0; // 0x8000
  if (neg1)
    IAnegate(x,count);
  var neg2 = (y[0] & SIGNMASK) != 0; // 0x8000
  if (neg2)
    IAnegate(y,count);
  if (x[0] == 0) {
    for (var i = 0; i < count; i++)
      multmp[i] = 0;
  } else if (x[0] == 1) {
    for (var i = 0; i < count; i++)
      multmp[i] = y[i];
  } else {
    var carry = 0;
    for (var i = count-1; i >= 0; i--) {
      multmp[i] = x[0]*y[i] + carry;
      carry =  (multmp[i] * IArecip) & FRACMASK; // multmp[i] >>> 16
      multmp[i] &= FRACMASK; // 0xFFFF
    }
  }
  for (var j = 1; j < count; j++) {
    var i = count - j;
    var carry = (x[j]*y[i] * IArecip) & FRACMASK; // (x[j]*y[i]) >>> 16
    i--;
    var k = count - 1;
    while (i >= 0) {
      multmp[k] += x[j]*y[i] + carry;
      carry = (multmp[k] * IArecip) & FRACMASK; // multmp[k] >>> 16
      multmp[k] &= FRACMASK; // 0xFFFF
      i--;
      k--;
    }
    while (carry != 0 && k >= 0) {
      multmp[k] += carry;
      carry = (multmp[k] * IArecip) & FRACMASK; // multmp[k] >>> 16
      multmp[k] &= FRACMASK; // 0xFFFF
      k--;
    }
  }
  IAcopy(x, 0, multmp, 0, count);
  if (neg2)
    IAnegate(y,count);
  if (neg1 != neg2)
    IAnegate(x,count);
} // End of IA.multiply

function dwellHP( /* Uint32Array */ r, /* Uint32Array */ i)
{
  var count = 0;
  IAcopy(qzr, 0, r, 0, chunks); // zr := r
  IAcopy(qzi, 0, i, 0, chunks); // zi := i
  var zr, zi; // used as temps in derivative iteration
  IAcopy(qzr2, 0, qzr, 0, chunks); // zr2 := zr
  IAmultiply(qzr2, qzr, chunks);   // zr2 = zr*zr
  IAcopy(qzi2, 0, qzi, 0, chunks); // zi2 := zi
  IAmultiply(qzi2, qzi, chunks);   // zi2 = zi*zi
  var dzr = 1.0;
  var dzi = 0.0;
  var rad2 = 0.0;
  var gg = true;
  while ((count < maxIterations) & gg) {
    zr = IA_f64(qzr, chunks); zi = IA_f64(qzi, chunks);
    rad2 = 2.0*(zr*dzr - zi*dzi) + 1.0;
    dzi = 2.0*(zi*dzr + zr*dzi);
    dzr = rad2;

    IAcopy(qzit, 0, qzr, 0, chunks);  // zit = zr
    IAadd(qzit, qzr, chunks);  // zit = 2*zr
    IAmultiply(qzit, qzi, chunks);  // zit = 2*zr*zi
    IAadd(qzit, i, chunks);  // zit = 2*zr*zi + i, next value for zi
    IAcopy(qzi, 0, qzit, 0, chunks);  // zi = 2*zr*zi + i

    // We use multmp to negate zi2 because there is no subtract
    IAcopy(multmp, 0, qzi2, 0, chunks);   // multmp = zi*zi
    IAnegate(multmp, chunks);  // multmp = -zi*zi
    IAadd(qzr2, multmp, chunks);  // zr2 = zr*zr - zi*zi
    IAadd(qzr2, r, chunks); // zr2 = zr*zr - zi*zi + r, next value for zr
    IAcopy(qzr, 0, qzr2, 0, chunks);   // zr = zr*zr - zi*zi + r

    IAcopy(qzr2, 0, qzr, 0, chunks); // zr2 := zr
    IAmultiply(qzr2, qzr, chunks);   // zr2 = zr*zr
    IAcopy(qzi2, 0, qzi, 0, chunks); // zi2 := zi
    IAmultiply(qzi2, qzi, chunks);   // zi2 = zi*zi

    count++;

    IAcopy(qrad2, 0, qzi2, 0, chunks); // rad2 = zi*zi
    IAadd(qrad2, qzr2, chunks);  // rad2 = zr*zr + zi*zi
    if (true) {
      // new code
      if (qrad2[0] >= escrad2) {
//      if (IA_escaped(qrad2[0], qrad2[1])) { // (qrad2[0] & 0xFFF8) == 0)
        gg = false;
      }
    } else {
      if (qrad2[0] >= 8) {
        gg = false;
      }
    }
    // else: positive and less than 8, keep iterating
  }
  if (count < maxIterations) {
    var zr2, zi2;
    rad2 = IA_f64(qrad2, chunks);
    zr2 = (dzr*dzr)+(dzi*dzi);
    if (zr2 > 0.0) {
      zi2 = 2.0 * Math.log(Math.sqrt(rad2)) * Math.sqrt(rad2/zr2);
    } else {
      zi2 = 1.0;
    }
    if (show_dem) { // if (show_dem)
      if (zi < 0.0) {
        count++;
      }
    }
    return [count, zi2];
  }
  /* else, not escaped */
  return [-1 , 1.0];
} // End of dwell.HP
